marimo_test_{{JOB_NAME}}:
  stage: test
  image: ghcr.io/astral-sh/uv:0.9.30-bookworm
  variables:
    GIT_LFS_SKIP_SMUDGE: "0"
  before_script:
    - apt-get update && apt-get install -y git-lfs
    - git lfs install
    - git lfs pull
  script:
    - echo "🚀 Running notebook: {{NOTEBOOK}}"
    - folder="results/$(basename "{{NOTEBOOK}}" .py)"
    - mkdir -p "$folder"
    - export NOTEBOOK_OUTPUT_FOLDER="$folder"
    - uv run "{{NOTEBOOK}}"
  artifacts:
    paths:
      - results/
    expire_in: 1 week
  allow_failure: false
